"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var clean_stderr_1 = require("./clean-stderr");
var JStackThreadState;
(function (JStackThreadState) {
    JStackThreadState["NEW"] = "NEW";
    JStackThreadState["RUNNABLE"] = "RUNNABLE";
    JStackThreadState["SLEEPING"] = "SLEEPING";
    JStackThreadState["IN_OBJECT_WAIT"] = "IN_OBJECT_WAIT";
    JStackThreadState["IN_OBJECT_WAIT_TIMED"] = "IN_OBJECT_WAIT_TIMED";
    JStackThreadState["PARKED"] = "PARKED";
    JStackThreadState["PARKED_TIMED"] = "PARKED_TIMED";
    JStackThreadState["BLOCKED_ON_MONITOR_ENTER"] = "BLOCKED_ON_MONITOR_ENTER";
    JStackThreadState["TERMINATED"] = "TERMINATED";
    JStackThreadState["TIMED_WAITING"] = "TIMED_WAITING";
    JStackThreadState["WAITING"] = "WAITING";
    JStackThreadState["UNKNOWN"] = "UNKNOWN";
})(JStackThreadState = exports.JStackThreadState || (exports.JStackThreadState = {}));
function parseJStack(output, stderr) {
    if (stderr) {
        stderr = clean_stderr_1.cleanStderr(stderr);
        if (stderr.indexOf('Operation not permitted') >= 0) {
            throw new Error('Operation not permitted');
        }
        else if (stderr.indexOf('No such process') >= 0) {
            throw new Error('No such process');
        }
        else if (stderr.length > 0) {
            throw new Error("Unhandled error: " + stderr);
        }
    }
    var lines = output.split('\n');
    var results = {
        threads: [],
    };
    for (var i = 0; i < lines.length;) {
        var line = lines[i];
        i++;
        var dateTimeMatch = line.match(/([0-9]+)-([0-9]+)-([0-9]+) ([0-9]+):([0-9]+):([0-9]+)/);
        if (dateTimeMatch) {
            var year = parseInt(dateTimeMatch[1], 10);
            var month = parseInt(dateTimeMatch[2], 10);
            var day = parseInt(dateTimeMatch[3], 10);
            var hour = parseInt(dateTimeMatch[4], 10);
            var minute = parseInt(dateTimeMatch[5], 10);
            var second = parseInt(dateTimeMatch[6], 10);
            results.dateTime = new Date(year, month - 1, day, hour, minute, second);
            continue;
        }
        var jdkMatch = line.match(/^Full thread dump (.*)$/);
        if (jdkMatch) {
            if (jdkMatch[1].endsWith(':')) {
                jdkMatch[1] = jdkMatch[1].substr(0, jdkMatch[1].length - 1);
            }
            results.jdk = jdkMatch[1];
            continue;
        }
        if (line.indexOf('Threads class SMR info') >= 0) {
            for (; i < lines.length; i++) {
                if (lines[i].trim().length === 0) {
                    break;
                }
            }
            continue;
        }
        var jniInfoMatch = line.match(/JNI global refs: ([0-9]+), weak refs: ([0-9]+)/);
        if (jniInfoMatch) {
            results.jniGlobalRefs = parseInt(jniInfoMatch[1], 10);
            results.jniWeakRefs = parseInt(jniInfoMatch[1], 10);
            continue;
        }
        var jniInfoAltMatch = line.match(/JNI global references: ([0-9]+)/);
        if (jniInfoAltMatch) {
            results.jniGlobalRefs = parseInt(jniInfoAltMatch[1], 10);
            continue;
        }
        var threadStartMatch = line.match(/^"(.*?)" (.*)$/);
        if (threadStartMatch) {
            var name_1 = threadStartMatch[1];
            var info = threadStartMatch[2];
            var thread = {
                name: name_1,
                daemon: info.indexOf('daemon') >= 0,
                stackItems: [],
            };
            info = info.replace('daemon', '');
            var javaThreadIdMatch = info.match(/#([0-9]+)/);
            if (javaThreadIdMatch) {
                info = info.replace(javaThreadIdMatch[0], '');
                thread.javaThreadId = parseInt(javaThreadIdMatch[1], 10);
            }
            var osPriorityMatch = info.match(/os_prio=([0-9]+)/);
            if (osPriorityMatch) {
                info = info.replace(osPriorityMatch[0], '');
                thread.osPriority = parseInt(osPriorityMatch[1], 10);
            }
            var priorityMatch = info.match(/prio=([0-9]+)/);
            if (priorityMatch) {
                info = info.replace(priorityMatch[0], '');
                thread.priority = parseInt(priorityMatch[1], 10);
            }
            var cpuTimeMatch = info.match(/cpu=([0-9.]+[a-z]+)/);
            if (cpuTimeMatch) {
                info = info.replace(cpuTimeMatch[0], '');
                thread.cpuTimeMillis = parseTimeToMillis(cpuTimeMatch[1]);
            }
            var elapsedTimeMatch = info.match(/elapsed=([0-9.]+[a-z]+)/);
            if (elapsedTimeMatch) {
                info = info.replace(elapsedTimeMatch[0], '');
                thread.elapsedTimeMillis = parseTimeToMillis(elapsedTimeMatch[1]);
            }
            var tidMatch = info.match(/tid=0x([0-9a-f]+)/);
            if (tidMatch) {
                info = info.replace(tidMatch[0], '');
                thread.tid = "0x" + tidMatch[1];
            }
            var nidMatch = info.match(/nid=0x([0-9a-f]+)/);
            if (nidMatch) {
                info = info.replace(nidMatch[0], '');
                thread.nid = parseInt(nidMatch[1], 16);
            }
            var stackMemoryRegionMatch = info.match(/\[0x([0-9a-f]+)]/);
            if (stackMemoryRegionMatch) {
                info = info.replace(stackMemoryRegionMatch[0], '');
                thread.stackMemoryRegion = "0x" + stackMemoryRegionMatch[1];
            }
            thread.info = info.replace(/\s+/, ' ').trim();
            var stateLineMatch = lines[i].match(/java.lang.Thread.State: (.*)/);
            if (stateLineMatch) {
                i++;
                thread.state = parseThreadState(stateLineMatch[1].trim());
            }
            for (; i < lines.length; i++) {
                var stackItemLine = lines[i].trim();
                if (stackItemLine.length === 0) {
                    break;
                }
                thread.stackItems.push({
                    line: stackItemLine,
                });
            }
            results.threads.push(thread);
            continue;
        }
        if (line.trim().length === 0) {
            continue;
        }
        throw new Error("Unhandled line: " + line + " (line " + i + ")");
    }
    return results;
}
exports.parseJStack = parseJStack;
function parseTimeToMillis(time) {
    if (time.endsWith('ms')) {
        return parseFloat(time.substr(0, time.length - 2));
    }
    if (time.endsWith('s')) {
        return parseFloat(time.substr(0, time.length - 1)) * 1000;
    }
    throw new Error("Could not parse time: " + time);
}
function parseThreadState(str) {
    if (str === 'NEW') {
        return JStackThreadState.NEW;
    }
    if (str === 'RUNNABLE') {
        return JStackThreadState.RUNNABLE;
    }
    if (str === 'TERMINATED') {
        return JStackThreadState.TERMINATED;
    }
    if (str === 'UNKNOWN') {
        return JStackThreadState.UNKNOWN;
    }
    if (str.startsWith('BLOCKED')) {
        return JStackThreadState.BLOCKED_ON_MONITOR_ENTER;
    }
    if (str.startsWith('TIMED_WAITING')) {
        if (str.indexOf('parking') >= 0) {
            return JStackThreadState.PARKED_TIMED;
        }
        if (str.indexOf('sleeping') >= 0) {
            return JStackThreadState.SLEEPING;
        }
        if (str.indexOf('on object monitor') >= 0) {
            return JStackThreadState.IN_OBJECT_WAIT_TIMED;
        }
        return JStackThreadState.TIMED_WAITING;
    }
    if (str.startsWith('WAITING')) {
        if (str.indexOf('on object monitor') >= 0) {
            return JStackThreadState.IN_OBJECT_WAIT;
        }
        if (str.indexOf('parking') >= 0) {
            return JStackThreadState.PARKED;
        }
        return JStackThreadState.WAITING;
    }
    throw new Error("Could not parse state: " + str);
}
//# sourceMappingURL=jstack-parse.js.map